IMPLEMENTATION MODULE TextBuffers;

        (********************************************************)
        (*                                                      *)
        (*        Text buffers for line-oriented file I/O       *)
        (*                                                      *)
        (*          (At present we support only input)          *)
        (*                                                      *)
        (*   This module maintains a read-ahead buffer for each *)
        (*   open file.  Obviously it will be efficient to do   *)
        (*   this only if input is mostly sequential.           *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            16 July 2001                    *)
        (*  Last edited:        6 November 2001                 *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT LOC, BYTE, CAST;

FROM Types IMPORT CARD64;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilePos,
    (* proc *)  OpenOldFile, CloseFile, StartPosition,
                EndPosition, SetPosition, ReadRaw;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM LONGLONG IMPORT
    (* proc *)  Sum64, Compare64, ShortSub;

(************************************************************************)

CONST
    NumberOfBuffers = 8;
    BufferSize = 4096;
    CtrlZ = CHR(26);

TYPE
    (* Each data buffer contains the following fields.                  *)
    (*    StartPos  the position within the file from which these data  *)
    (*              were read                                           *)
    (*    count     the number of bytes of data                         *)
    (*    data      the actual cached data                              *)

    BufferPointer = POINTER TO
                        RECORD
                            StartPos: FilePos;
                            count: CARDINAL;
                            data: ARRAY [0..BufferSize-1] OF CHAR;
                        END (*RECORD*);

    (* A Buffer record contains NumberOfBuffers data buffers, each of   *)
    (* which is a cached section of the currently opened file.  The     *)
    (* fields are                                                       *)
    (*    cid            channel ID of the opened file                  *)
    (*    CurrentPosition  the location in the file from which the      *)
    (*                     next character will come, assuming that      *)
    (*                     we continue to read sequentially             *)
    (*    CurrentBuffer  the buffer from which we're currently reading  *)
    (*                   data.  If this is zero, we need to switch to   *)
    (*                   a new buffer.                                  *)
    (*    BufPtr         points to an array of buffers                  *)

    Buffer = POINTER TO
                 RECORD
                     cid: ChanId;
                     CurrentPosition: FilePos;
                     CurrentBuffer: [0..NumberOfBuffers];
                     BufPtr: ARRAY [1..NumberOfBuffers] OF BufferPointer;
                 END (*RECORD*);

(************************************************************************)
(*                           OPEN/CLOSE/ETC                             *)
(************************************************************************)

PROCEDURE OpenForReading (name: ARRAY OF CHAR): Buffer;

    (* Opens an existing file for read-only access, and returns its     *)
    (* buffer ID.                                                       *)

    VAR result: Buffer;  k: CARDINAL;

    BEGIN
        NEW (result);
        WITH result^ DO
            cid := OpenOldFile (name, FALSE);
            FOR k := 1 TO NumberOfBuffers DO
                NEW (BufPtr[k]);
                WITH BufPtr[k]^ DO
                    StartPos := CARD64{0,0};  count := 0;
                END (*WITH*);
            END (*FOR*);
            CurrentBuffer := 1;
            IF cid <> NoSuchChannel THEN
                CurrentPosition := StartPosition(cid);
                ReadRaw (cid, BufPtr[1]^.data, BufferSize, BufPtr[1]^.count);
            END (*IF*);
            IF BufPtr[1]^.count = 0 THEN
                CurrentBuffer := 0;
            END (*IF*);
        END (*WITH*);
        RETURN result;
    END OpenForReading;

(************************************************************************)

PROCEDURE TBFileOpened (TB: Buffer): BOOLEAN;

    (* Use this to check whether an 'open' operation succeeded. *)

    BEGIN
        RETURN (TB <> NIL) AND (TB^.cid <> NoSuchChannel);
    END TBFileOpened;

(************************************************************************)

PROCEDURE CloseTB (VAR (*INOUT*) TB: Buffer);

    (* Closes a file. *)

    VAR k: CARDINAL;

    BEGIN
        IF TB <> NIL THEN
            CloseFile (TB^.cid);
            FOR k := 1 TO NumberOfBuffers DO
                DISPOSE (TB^.BufPtr[k]);
            END (*FOR*);
            DISPOSE (TB);
        END (*IF*);
    END CloseTB;

(************************************************************************)
(*                         LOADING THE DATA                             *)
(************************************************************************)

PROCEDURE LoadNextBuffer (TB: Buffer);

    (* Reads some more data from the file.  If TB^.CurrentBuffer = 0    *)
    (* after this operation, we must have reached the end of the file.  *)

    BEGIN
        WITH TB^ DO
            IF cid <> NoSuchChannel THEN
                SetPosition (cid, CurrentPosition);
                CurrentBuffer := NumberOfBuffers;
                WITH BufPtr[NumberOfBuffers]^ DO
                    StartPos := CurrentPosition;
                    ReadRaw (cid, data, BufferSize, count);
                END (*WITH*);
            END (*IF*);
            IF BufPtr[NumberOfBuffers]^.count = 0 THEN
                CurrentBuffer := 0;
            END (*IF*);
        END (*WITH*);
    END LoadNextBuffer;

(************************************************************************)

PROCEDURE SetCurrentBuffer (TB: Buffer);

    (* Updates TB^.CurrentBuffer to be a buffer whose data includes     *)
    (* a character at TB^.CurrentPosition.  This might require reading  *)
    (* more data from the file.  If TB^.CurrentBuffer = 0 after this    *)
    (* operation, we must have reached the end of the file.             *)

    VAR k: CARDINAL;

    BEGIN
        k := 1;
        LOOP
            IF k > NumberOfBuffers THEN
                LoadNextBuffer (TB);
                EXIT (*LOOP*);
            ELSE
                WITH TB^ DO
                    WITH BufPtr[k]^ DO
                        IF (Compare64(CurrentPosition, StartPos) >= 0)
                                AND (Compare64(CurrentPosition,
                                         Sum64(StartPos, CARD64{count,0})) < 0) THEN
                            CurrentBuffer := k;
                            EXIT (*LOOP*);
                        END (*IF*);
                    END (*WITH*);
                END (*WITH*);
                INC (k);
            END (*IF*);
        END (*LOOP*);
    END SetCurrentBuffer;

(************************************************************************)
(*                         FILE POSITION/SIZE                           *)
(************************************************************************)

PROCEDURE TBCurrentPosition (TB: Buffer): FilePos;

    (* Returns the current position within the file. *)

    BEGIN
        RETURN TB^.CurrentPosition;
    END TBCurrentPosition;

(************************************************************************)

PROCEDURE TBStartPosition (TB: Buffer): FilePos;

    (* Returns the start-of-file position. *)

    BEGIN
        RETURN StartPosition (TB^.cid);
    END TBStartPosition;

(************************************************************************)

PROCEDURE TBEndPosition (TB: Buffer): FilePos;

    (* Returns the end-of-file position. *)

    BEGIN
        RETURN EndPosition (TB^.cid);
    END TBEndPosition;

(************************************************************************)

PROCEDURE TBSetPosition (TB: Buffer;  position: FilePos);

    (* Sets the current position within the file. *)

    BEGIN
        SetPosition (TB^.cid, position);
        TB^.CurrentPosition := position;
        SetCurrentBuffer (TB);
    END TBSetPosition;

(************************************************************************)
(*                              INPUT                                   *)
(************************************************************************)

(************************************************************************)
(*   Remark: the following procedures could be cleaned up a lot.        *)
(************************************************************************)

PROCEDURE ReadByte (TB: Buffer;  VAR (*OUT*) value: BYTE): BOOLEAN;

    (* Reads a single character, returns FALSE if end-of-file. *)

    BEGIN
        IF TB^.CurrentBuffer = 0 THEN
            SetCurrentBuffer (TB);
            IF TB^.CurrentBuffer = 0 THEN
                RETURN FALSE;
            END (*IF*);
        END (*IF*);
        WITH TB^ DO
            WITH BufPtr[CurrentBuffer]^ DO
                value := CAST(BYTE, data[ShortSub(CurrentPosition, StartPos)]);
            END (*WITH*);
            IF CurrentPosition.low = MAX(CARDINAL) THEN
                CurrentPosition.low := 0;
                INC (CurrentPosition.high);
            ELSE
                INC (CurrentPosition.low);
            END (*IF*);
            IF Compare64(CurrentPosition,
                    Sum64(BufPtr[CurrentBuffer]^.StartPos,
                          CARD64{BufPtr[CurrentBuffer]^.count,0})) >= 0 THEN
                CurrentBuffer := 0;
            END (*IF*);
        END (*WITH*);
        RETURN TRUE;
    END ReadByte;

(************************************************************************)

PROCEDURE TBReadRaw (TB: Buffer;  VAR (*OUT*) data: ARRAY OF LOC;
                   limit: CARDINAL;  VAR (*OUT*) NumberRead: CARDINAL);

    (* Reads a buffer-full of information from a file. *)

    VAR j: CARDINAL;
        temp: BufferPointer;

    BEGIN
        j := 0;
        WHILE (j < limit) AND (j <= HIGH(data))
                          AND ReadByte (TB, data[j]) DO
            INC(j);
        END (*WHILE*);

        NumberRead := j;

        (* Promote current buffer as an approximate LRU calculation. *)

        WITH TB^ DO
            IF CurrentBuffer > 1 THEN
                temp := BufPtr[CurrentBuffer-1];
                BufPtr[CurrentBuffer-1] := BufPtr[CurrentBuffer];
                BufPtr[CurrentBuffer] := temp;
                DEC (CurrentBuffer);
            END (*IF*);
        END (*WITH*);

    END TBReadRaw;

(************************************************************************)

PROCEDURE ReadChar (TB: Buffer;  VAR (*OUT*) ch: CHAR);

    (* Reads a single character, returns CtrlZ if end-of-file. *)

    BEGIN
        IF TB^.CurrentBuffer = 0 THEN
            SetCurrentBuffer (TB);
            IF TB^.CurrentBuffer = 0 THEN
                ch := CtrlZ;
                RETURN;
            END (*IF*);
        END (*IF*);
        WITH TB^ DO
            WITH BufPtr[CurrentBuffer]^ DO
                ch := data[ShortSub(CurrentPosition, StartPos)];
            END (*WITH*);
            IF CurrentPosition.low = MAX(CARDINAL) THEN
                CurrentPosition.low := 0;
                INC (CurrentPosition.high);
            ELSE
                INC (CurrentPosition.low);
            END (*IF*);
            IF Compare64(CurrentPosition,
                    Sum64(BufPtr[CurrentBuffer]^.StartPos,
                          CARD64{BufPtr[CurrentBuffer]^.count,0})) >= 0 THEN
                CurrentBuffer := 0;
            END (*IF*);
        END (*WITH*);
    END ReadChar;

(************************************************************************)

PROCEDURE TBReadLine (TB: Buffer;  VAR (*OUT*) data: ARRAY OF CHAR);

    (* Reads a line of text from a file.  Assumption: a line ends with  *)
    (* CRLF.  To avoid tortuous logic, I take the CR as end of line and *)
    (* skip the LF.  At end of file we return with data[0] = Ctrl/Z.    *)

    CONST Nul = CHR(0);  CR = CHR(13);  LF = CHR(10);

    VAR j: CARDINAL;
        ch: CHAR;
        temp: BufferPointer;

    BEGIN
        j := 0;  ch := Nul;
        LOOP
            ReadChar (TB, ch);
            IF ch = CR THEN
                (* ignore carriage return. *)
            ELSIF ch = LF THEN
                EXIT (*LOOP*);
            ELSIF j <= HIGH(data) THEN
                data[j] := ch;  INC(j);
                IF ch = CtrlZ THEN
                    EXIT (*LOOP*);
                END (*IF*);
            END (*IF*);
        END (*LOOP*);

        IF j <= HIGH(data) THEN
            data[j] := Nul;
        END (*IF*);

        (* Promote current buffer as an approximate LRU calculation. *)

        WITH TB^ DO
            IF CurrentBuffer > 1 THEN
                temp := BufPtr[CurrentBuffer-1];
                BufPtr[CurrentBuffer-1] := BufPtr[CurrentBuffer];
                BufPtr[CurrentBuffer] := temp;
                DEC (CurrentBuffer);
            END (*IF*);
        END (*WITH*);

    END TBReadLine;

(************************************************************************)

END TextBuffers.

